/* gutil1.c: Some graph utilities. */

#define EXTDEFS 1

#include "gtools.h"
#include "gutils.h"

/**************************************************************************/

void
degstats(graph *g, int m, int n, unsigned long *edges,
         int *mindeg, int *mincount, int *maxdeg, int *maxcount,
         boolean *eulerian)
/* Compute degree-related graph properties.
   *edges = number of edges
   *mindeg, *mincount = minimum degree and how many there are
   *maxdeg, *maxcount = maximum degree and how many there are
   *eulerian = whether the graph has only even degrees
*/
{
	setword *pg;
	int i,j,d,dor;
	int mind,mindc,maxd,maxdc;
	unsigned long ned;

	mind = n;  mindc = 0;
	maxd = 0;  maxdc = 0;
	ned = 0;
        dor = 0;

	pg = (setword*)g;
	for (i = 0; i < n; ++i)
	{
	    d = 0;
	    for (j = 0; j < m; ++j, ++pg)
	        if (*pg) d += POPCOUNT(*pg);

	    if (d == mind)
	 	++mindc;
	    else if (d < mind)
	    {
		mind = d;
		mindc = 1;
	    }

	    if (d == maxd)
	 	++maxdc;
	    else if (d > maxd)
	    {
		maxd = d;
		maxdc = 1;
	    }

	    dor |= d;
	    ned += d;
	}

	*mindeg = mind;
	*mincount = mindc;
	*maxdeg = maxd;
	*maxcount = maxdc;
	*edges = ned / 2;
	*eulerian = (dor & 1) == 0;
}

/*********************************************************************/

boolean
isconnected1(graph *g, int n)
/* test if g is connected (m=1) */
{
        register setword seen,expanded,toexpand;
        register int i;

        seen = bit[0];
        expanded = 0;

        while (toexpand = (seen & ~expanded))             /* not == */
        {
            i = FIRSTBIT(toexpand);
            expanded |= bit[i];
            seen |= g[i];
        }

        return  POPCOUNT(seen) == n;
}

/**************************************************************************/

boolean
isconnected(graph *g, int m, int n)
/* Test if g is connected */
{
	int i,head,tail,w;
	set *gw;
#if MAXN
        nvector queue[MAXN],visited[MAXN];
#else
        DYNALLSTAT(nvector,queue,queue_sz);
        DYNALLSTAT(nvector,visited,visited_sz);
#endif

	if (m == 1) return isconnected1(g,n);

#if !MAXN
        DYNALLOC1(nvector,queue,queue_sz,n,"isconnected");
        DYNALLOC1(nvector,visited,visited_sz,n,"isconnected");
#endif

	for (i = 0; i < n; ++i)
	    visited[i] = 0;

	queue[0] = 0;
        visited[0] = 1;

	head = 0;
	tail = 1;
	while (head < tail)
	{
	    w = queue[head++];
            gw = GRAPHROW(g,w,m);
            for (i = -1; (i = nextelement(gw,m,i)) >= 0;)
            {
                if (!visited[i])
                {
                    visited[i] = 1;
                    queue[tail++] = i;
                }
            }
        }

	return tail == n;
}

/**************************************************************************/

boolean
isbipartite(graph *g, int m, int n)
/* Test if g is bipartite */
{
        int i,head,tail,v,w,need;
        set *gw;
#if MAXN
        nvector queue[MAXN],colour[MAXN];
#else
        DYNALLSTAT(nvector,queue,queue_sz);
        DYNALLSTAT(nvector,colour,colour_sz);
#endif

        /* if (m == 1) return isbipartite1(g,n);  */

#if !MAXN
        DYNALLOC1(nvector,queue,queue_sz,n,"isbipartite");
        DYNALLOC1(nvector,colour,colour_sz,n,"isbipartite");
#endif

        for (i = 0; i < n; ++i)
            colour[i] = -1;

	for (v = 0; v < n; ++v)
	if (colour[v] < 0)
	{
            queue[0] = v;
            colour[v] = 0;

            head = 0;
            tail = 1;
            while (head < tail) 
            {
                w = queue[head++];
		need = 1 - colour[w];
                gw = GRAPHROW(g,w,m);
                for (i = -1; (i = nextelement(gw,m,i)) >= 0;)
                {
                    if (colour[i] < 0)
                    {
                        colour[i] = need;
                        queue[tail++] = i;
                    }
		    else if (colour[i] != need)
			return FALSE;
                }
            }
	}

	return TRUE;
}

/**************************************************************************/

void
find_dist(graph *g, int m, int n, int v, int *dist)
/* Put in dist[0..n-1] the distance of each vertex from v.
   Vertices in a different component are given the distance n. */
{
        int i,head,tail,w;
        set *gw;
#if MAXN
        nvector queue[MAXN];
#else   
        DYNALLSTAT(nvector,queue,queue_sz);
#endif  
        
#if !MAXN
        DYNALLOC1(nvector,queue,queue_sz,n,"isconnected");
#endif  
        
        for (i = 0; i < n; ++i)
            dist[i] = n;

        queue[0] = v;
        dist[v] = 0;

        head = 0;
        tail = 1;
        while (tail < n && head < tail)
        {
            w = queue[head++];
            gw = GRAPHROW(g,w,m);
            for (i = -1; (i = nextelement(gw,m,i)) >= 0;)
            {
                if (dist[i] == n)
                {
                    dist[i] = dist[w] + 1;
                    queue[tail++] = i;
                }
            }
        }
}

/**************************************************************************/

void
diamstats(graph *g, int m, int n, int *radius, int *diameter)
/* Find the radius and diameter.  Both -1 if g is disconnected.
   We use an O(mn) algorithm, which is pretty disgraceful. */
{
	int v,i,head,tail,w;
	int ecc,diam,rad;
	set *gw;
#if MAXN
        nvector queue[MAXN],dist[MAXN];
#else
        DYNALLSTAT(nvector,queue,queue_sz);
        DYNALLSTAT(nvector,dist,dist_sz);
#endif

	/* if (m == 1) {diamstats1(g,n,radius,diameter); return; } */

#if !MAXN
        DYNALLOC1(nvector,queue,queue_sz,n,"isconnected");
        DYNALLOC1(nvector,dist,dist_sz,n,"isconnected");
#endif

	diam = -1;
	rad = n;

	for (v = 0; v < n; ++v)
	{
	    for (i = 0; i < n; ++i)
	        dist[i] = -1;

	    queue[0] = v;
            dist[v] = 0;

	    head = 0;
	    tail = 1;
	    while (tail < n && head < tail)
	    {
	        w = queue[head++];
                gw = GRAPHROW(g,w,m);
                for (i = -1; (i = nextelement(gw,m,i)) >= 0;)
                {
                    if (dist[i] < 0)
                    {
                        dist[i] = dist[w] + 1;
                        queue[tail++] = i;
                    }
                }
            }

	    if (tail < n)
	    {
		*diameter = *radius = -1;
		return;
	    }

	    ecc = dist[queue[n-1]];

	    if (ecc > diam) diam = ecc;
	    if (ecc < rad)  rad  = ecc;
	}

	*diameter = diam;
	*radius = rad;
}
