/* splay.c  - code for splay trees

   This file is not meant to be compiled separately, but to be
   #included into other programs.  Use it like this:

   1. Define a node type SPLAYNODE.  It must be a structure that
   contains at least the pointer fields left, right and parent.

   2. Declare a variable of type SPLAYNODE* to point to the root
   of the tree, and initialise it to NULL.

   3. Declare SCAN_ARGS to be the additional arguments needed for
   splay_scan(), including a leading comma.

   4. Declare ACTION(p) for what splay_scan() should do for node p.

   5. Declare INSERT_ARGS to be the additional arguments needed 
   for splay_insert(), including a leading comma.

   6. Declare COMPARE(p) to compare INSERT_ARGS to the contents of
   node p.  <0, 0, >0 if INSERT_ARGS is greater, equal, less, than p.
   
   7. Declare PRESENT(p) for what to do if INSERT_ARGS is already
   present, in node p.  There is a spare int variable i available.

   8. Declare NOT_PRESENT(p) for what to do if INSERT_ARGS is not
   in the tree and p is a fresh node to hold it.  No need to set
   the left, right, and parent fields.  Use i here too if you like.

	PRESENT() and NOT_PRESENT() should not manipulate the
        tree pointers.  However, each of them can include a
        return if you don't want to change the tree.  In the
        case of NOT_PRESENT(), do free(p) before returning.

   9. #include "splay.c"
*/

#define S_A 0
#define S_L 1
#define S_R 2
#define S_LL 3
#define S_LR 4
#define S_RL 5
#define S_RR 6

/*********************************************************************/

static void
splay_scan(SPLAYNODE *root SCAN_ARGS)
/* Do ACTION(p) for each node of the tree, in inorder.  Nonrecursive! */
{
    int code;
    SPLAYNODE *p;

    p = root;
    code = S_A;

    while (p)
    {
	switch (code)    /* deliberate flow-ons */
	{
	 case S_A:
	    if (p->left)
	    {
		p = p->left;
		break;
	    }
	 case S_L:
	    ACTION(p);
	    if (p->right)
	    {
		p = p->right;
		code = S_A;
		break;
	    }
	 case S_R:
	    if (p->parent && p->parent->left == p) code = S_L;
	    else                                   code = S_R;
	    p = p->parent;
	    break;
	}
    }
}

/*********************************************************************/

static void
splay(SPLAYNODE *p)
/* Splay the node p.  It becomes the new root. */
{
    SPLAYNODE *q,*r,*s;
    SPLAYNODE *a,*b,*c,*d;
    int code;

#define LCHILD(x,y) {(x)->left = y; if (y) (y)->parent = x;}
#define RCHILD(x,y) {(x)->right = y; if (y) (y)->parent = x;}

    while (p->parent)
    {
	a = p->left;
	b = p->right;
	q = p->parent;
	if (q->left == p)
	{
	    code = S_L;
	    c = q->right;
	}
	else
	{
	    code = S_R;
	    c = q->left;
	}
	r = q->parent;
	if (r)
	{
	    if (r->left == q)
	    {
		code = (code == S_L ? S_LL : S_LR);
		d = r->right;
	    }
	    else
	    {
		code = (code == S_L ? S_RL : S_RR);
		d = r->left;
	    }
	    s = r->parent;
	    p->parent = s;
	    if (s)
	    {
		if (s->left == r) s->left = p;
		else              s->right = p;
	    }
	}
	else
	{
	    p->parent = NULL;
	}
	
	switch (code)
	{
	 case S_L:
	    RCHILD(p,q);
	    LCHILD(q,b);
	    break;
	 case S_R:
	    LCHILD(p,q);
	    RCHILD(q,a);
	    break;
	 case S_LL:
	    RCHILD(p,q);
	    RCHILD(q,r);
	    LCHILD(q,b);
	    LCHILD(r,c);
	    break;
	 case S_RR:
	    LCHILD(p,q);
	    LCHILD(q,r);
	    RCHILD(r,c);
	    RCHILD(q,a);
	    break;
	 case S_LR:
	    LCHILD(p,q);
	    RCHILD(p,r);
	    RCHILD(q,a);
	    LCHILD(r,b);
	    break;
	 case S_RL:
	    LCHILD(p,r);
	    RCHILD(p,q);
	    RCHILD(r,a);
	    LCHILD(q,b);
	    break;
	}
    }
}

/*********************************************************************/

static void
splay_insert(SPLAYNODE **to_root  INSERT_ARGS)
/* Do insertion operation */
{
    int i,cmp;
    SPLAYNODE *p,*ppar,*new_node;

    p = *to_root;
    cmp = 0;

    while (p != NULL)
    {
        cmp = COMPARE(p);
        if (cmp == 0)
        {
	    PRESENT(p);
	    splay(p);
	    *to_root = p;
            return;
        }
        else if (cmp < 0)
        {
            ppar = p;
            p = p->left;
        }
        else
        {
            ppar = p;
            p = p->right;
        }
    }

    if ((new_node = (SPLAYNODE*)malloc(sizeof(SPLAYNODE))) == NULL)
    {
        fprintf(stderr,">E malloc failed in add_one()\n");
        exit(1);
    }

    NOT_PRESENT(new_node);

    new_node->left = new_node->right = NULL;

    if (cmp == 0)
    {
        *to_root = new_node;
	new_node->parent = NULL;
    }
    else if (cmp < 0)
    {
        ppar->left = new_node;
	new_node->parent = ppar;
    }
    else
    {
        ppar->right = new_node;
	new_node->parent = ppar;
    }

    splay(new_node);
    *to_root = new_node;
}
