/* sumlines.c - total the numbers appearing in various input lines.

   Usage:  sumlines [-fmt +fmt]... file file file ...
   Any number of input files can be given.  "-" means stdin.
   If there are no files given, just stdin is assumed.

   Extra or replacement formats can be given on the command line.
   -fmt replaces all previously-existing formats (except previous -fmt's).
   +fmt adds to the previously-existing formats.
   This means that +fmtA -fmtB is the same as -fmtB!
   A newline \n is assumed on the end of such formats.

   Example:   sumlines "-%d sticks" "-%d stones" file1 file2
*/

#define FINAL     1  /* At least one of these must be in each input file */
#define ERROR     2  /* Must be none of these */
#define UNIQUE    4  /* The %s and %c parts must be unique over all inputs */
#define COUNT     8  /* The output only states how many lines matched */
#define CONTINUE 16  /* Try to match later formats too */

static struct fmt_st
{
   int flags;
   char *fmt;
} format[] = {
   {ERROR,  ">E%s\n"},
   {ERROR,  ">Z -%s\n"},
   {ERROR,  ">C -%s\n"},
   {UNIQUE|COUNT, "plantri%s\n"},
   {UNIQUE|COUNT, "xplantri%s\n"},
   {UNIQUE|COUNT, ">A%s\n"},
   {0,      "%d rooted %s\n"},
   {0,      "Number with trivial group: %d\n"},
   {0,      " With %s edges and %s faces: %d\n"},
   {0,      " With %s: %d (%d classes)\n"},
   {0,      " With %s: %d\n"},
   {0,      " 2%s %d\n"},
   {0,      " 3%s %d\n"},
   {0,      " 4%s %d\n"},
   {0,      " 5%s %d\n"},
   {0,      " %s-%s : %d\n"},
   {0,      "Bipartite : %d\n"},
   {FINAL,  "%d d%sk triangulations generated; cpu=%f sec\n"},
   {FINAL,  "%d eulerian triangulations generated; cpu=%f sec\n"},
   {FINAL,  "%d %s generated (%d 1-conn; %d 2-conn; %d 3-conn)\n"},
   {FINAL,  "%d %s generated (%d 2-conn; %d 3-conn)\n"},
   {FINAL,  "%d 1-conn; %d 2-conn; %d 3-conn: %d %s generated; cpu=%f sec\n"},
   {FINAL,  "%d %s generated; cpu=%f sec\n"},
   {FINAL,  "%d %s written to %*s; cpu=%f sec\n"},
   {0,      "%d isomorphism classes; %d reflection invariant\n"},
   {FINAL,  "%d %sdiagrams with %s vertices and %s edges\n"},
   {0,      ">A geng%*s\n"},
   {CONTINUE,">C %d graphs with %s edges\n"},
   {0,      ">C %d graphs%*s\n"},
   {FINAL,  ">Z %d %s generated in %f sec%s\n"},
   {FINAL,  ">Z %d %s generated; cpu=%f\n"},
   {FINAL,  ">Z %d %s generated; %f sec%s\n"},
   {0,      "%d bipartite cubic graphs written to %*s\n"},
   {0,      "%d graphs read from %*s\n"},
   {FINAL,  "%d non-inout graphs written to %*s; %f sec\n"},
   {FINAL,  ">Z %d graphs from %*s, %d to %s, in %f sec.\n"},
   {FINAL,  ">Z %d graphs copied from %*s to %*s\n"},
   {0,      "%d graphs read from %*s\n"},
   {FINAL,  "%d nonhamiltonian graphs written to %*s; %f sec\n"},
   {UNIQUE|COUNT, "Case split_level=%s: %s modulo %s\n"},
   {0,      "Number of subsquares generated: %d\n"},
   {0,      ">Z %d squares read from %*s; %d productive\n"},
   {FINAL,  "%d triples written to %*s; %f sec.\n"},
   {0,      "cpu=%fu,%fs,%fx  pf=%d\n"}
};

/* The formats are tried against each input line one at a time, and the
   first one that matches is accepted.  The entire line must match.
   If the CONTINUE flag is present, the input line is also matched
   against further formats.

   Except in the case of formats with the COUNT flag, each format that
   matches any lines produces output giving the total value of each of the
   integers %d or real numbers %f in the lines which match.  If there are
   any %s or %c controls in the format, the output is given separately for
   each value of the matching strings which appear in the input lines.

   In the case of the COUNT flag, the program only reports the number of
   input lines which matched the format.

   If a format has the UNIQUE flag, no two input lines may match with the
   same values of the %s and %c controls present.  Otherwise a warning
   message is written for each duplicate match.

   %d  - matches an unsigned integer
   %f  - matches an unsigned real number of the form ddddd.ddd
   %sx - matches a string, where 'x' is any character. 
	 If 'x' is not a space, match zero or more characters from the
             current position up but not including the first 'x'.
         If 'x' is a space, match one or more characters from the current
             position up to and including the first non-space character
             which is followed by a space.
   %c  - matches a non-white character
   %%  - matches the character '%'
   %   - (with a space following the '%') matches zero or more spaces or
            tabs, as many as appear in the input.  In the output, this
            sequence appears as one space.
   %*d, %*f, %*sx, %*c - these are similar to the versions without the '*'
         except that the value is ignored (not used for summing, and not
         used to divide the output).  In the output, this field appears
         as a single character '*'.

  At least one FINAL format must match in each file or a warning is given.
*/

#define HAS(i,flgs)  ((format[i].flags&(flgs)) != 0)

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifdef __alpha
typedef long integer;
#define DOUT "%ld"       /* Formats used to output %d and %f quantities */
#define FOUT "%.2f"
#else
#ifdef __sun
typedef long long integer;
#define DOUT "%lld"      /* Formats used to output %d and %f quantities */
#define FOUT "%.2f"
#else
typedef long long integer;
#define DOUT "%Ld"       /* Formats used to output %d and %f quantities */
#define FOUT "%.2f"
#endif
#endif

static integer maxint;   /* set by find_maxint */

#define INCR(x,inc) \
     {if (maxint-(x) < (inc)) {fprintf(stderr,">E overflow!\n"); exit(1);} \
      x += (inc);}    /*  x += inc   with safety check */

typedef int boolean;
#define FALSE 0
#define TRUE  1

typedef union
{
    double f;
    integer d;
} number;

#define D 0    /* Code for "integer" */
#define F 1    /* Code for "real" */

#define MAXFORMATS 1000    /* A bound on NUMFORMATS */
#define NUMFORMATS (sizeof(format)/sizeof(struct fmt_st))
#define MAXLINELEN 10000   /* Maximum input line size
			    (longer lines are broken in two) */
#define MAXVALUES 16  /* Maximum total %d or %f items in a format */

typedef struct countrec
{
    struct countrec *left,*right,*parent;
    char *fmt;
    long count;
    number total[MAXVALUES];
} countnode;

static countnode *count_root[MAXFORMATS];
static long matching_lines[MAXFORMATS];

#define A 0
#define L 1
#define R 2
#define LL 3
#define LR 4
#define RL 5
#define RR 6

/****************************************************************************/

static void
writeline(char *outf, number *val)
/* Write an output line with the given format and values */
{
	int n;

	n = 0;

	for (; *outf != '\0'; ++outf)
	{
	    if (*outf == '%')
	    {
		++outf;
		if (*outf == '%')
		    putchar('%');
		else if (*outf == 'd')
		    printf(DOUT,val[n++].d);
		else if (*outf == 'f')
		    printf(FOUT,val[n++].f);
		else
		{
		    fprintf(stderr,">E unknown output format %%%c\n",*outf);
		    exit(1);
		}
	    }
	    else
		putchar(*outf);
	}
}

/*********************************************************************/

static void
print_counts(countnode *root)
/* Use a non-recursive inorder traversal to print the tree */
{
    int code;
    countnode *p;

    p = root;
    code = A;

    while (p)
    {
	switch (code)    /* deliberate flow-ons */
	{
	 case A:
	    if (p->left)
	    {
		p = p->left;
		break;
	    }
	 case L:
	    printf("%5ld: ",p->count);
	    writeline(p->fmt,p->total);
	    if (p->right)
	    {
		p = p->right;
		code = A;
		break;
	    }
	 case R:
	    if (p->parent && p->parent->left == p) code = L;
	    else                                   code = R;
	    p = p->parent;
	    break;
	}
    }
}

/*********************************************************************/

static void
print_common(countnode *root)
/* Print the common ends of the formats in the tree */
{
    int code;
    countnode *p;
    char *s0,*s1,*t0,*t1;
    int i,comm0,comm1,minlen,maxlen;

    if (root == NULL) return;

    p = root;
    code = A;

    s0 = s1 = p->fmt;
    while (*s1 != '\0') ++s1;
    comm0 = comm1 = minlen = maxlen = s1-s0;

    while (p)
    {   
        switch (code)    /* deliberate flow-ons */
        {
         case A:
            if (p->left)
            {   
                p = p->left;
                break;
            }
         case L:
	    t0 = t1 = p->fmt;
	    for (i = 0; i < comm0; ++i)
		if (s0[i] != t0[i]) break;
	    comm0 = i;

	    while (*t1 != '\0') ++t1;
	    for (i = 1; i <= comm1; ++i)
		if (s1[-i] != t1[-i]) break;
	    comm1 = i-1;
	    if (t1-t0 < minlen) minlen = t1-t0;
	    if (t1-t0 > maxlen) maxlen = t1-t0;

            if (p->right)
            {
                p = p->right;
                code = A;
                break;
            }
         case R:
            if (p->parent && p->parent->left == p) code = L;
            else                                   code = R;
            p = p->parent;
            break;
        }
    }

    if (comm0 + comm1 > minlen) comm1 = minlen - comm0;

    for (i = 0; i < comm0; ++i)
	printf("%c",s0[i]);
    if (comm0 + comm1 < maxlen) printf("*");
    for (i = comm1; i > 0; --i)
	printf("%c",s1[-i]);
}

/*********************************************************************/

static void
splay(countnode *p)
/* Splay the node p.  It becomes the new root. */
{
    countnode *q,*r,*s;
    countnode *a,*b,*c,*d;
    int code;

#define LCHILD(x,y) {(x)->left = y; if (y) (y)->parent = x;}
#define RCHILD(x,y) {(x)->right = y; if (y) (y)->parent = x;}

    while (p->parent)
    {
	a = p->left;
	b = p->right;
	q = p->parent;
	if (q->left == p)
	{
	    code = L;
	    c = q->right;
	}
	else
	{
	    code = R;
	    c = q->left;
	}
	r = q->parent;
	if (r)
	{
	    if (r->left == q)
	    {
		code = (code == L ? LL : LR);
		d = r->right;
	    }
	    else
	    {
		code = (code == L ? RL : RR);
		d = r->left;
	    }
	    s = r->parent;
	    p->parent = s;
	    if (s)
	    {
		if (s->left == r) s->left = p;
		else              s->right = p;
	    }
	}
	else
	{
	    p->parent = NULL;
	}
	
	switch (code)
	{
	 case L:
	    RCHILD(p,q);
	    LCHILD(q,b);
	    break;
	 case R:
	    LCHILD(p,q);
	    RCHILD(q,a);
	    break;
	 case LL:
	    RCHILD(p,q);
	    RCHILD(q,r);
	    LCHILD(q,b);
	    LCHILD(r,c);
	    break;
	 case RR:
	    LCHILD(p,q);
	    LCHILD(q,r);
	    RCHILD(r,c);
	    RCHILD(q,a);
	    break;
	 case LR:
	    LCHILD(p,q);
	    RCHILD(p,r);
	    RCHILD(q,a);
	    LCHILD(r,b);
	    break;
	 case RL:
	    LCHILD(p,r);
	    RCHILD(p,q);
	    RCHILD(r,a);
	    LCHILD(q,b);
	    break;
	}
    }
}

/*********************************************************************/

static void
add_one(countnode **to_root, char *fmt, int nval,
        number *val, int *valtype, int which)
/* Add 1 to the node with the given format, creating it if it is new.
   The tree is then splayed to maintain good efficiency. */
{
    int i,cmp,len;
    countnode *p,*ppar,*new_node;

    p = *to_root;
    cmp = 0;

    while (p != NULL)
    {
        cmp = strcmp(fmt,p->fmt);
        if (cmp == 0)
        {
	    if (HAS(which,UNIQUE) && p->count == 1)
		printf("ERROR: Multiple matches for %s",fmt);
            for (i = 0; i < nval; ++i)
		if (valtype[i] == D) {INCR(p->total[i].d,val[i].d);}
		else                 p->total[i].f += val[i].f;
	    ++p->count;
	    splay(p);
	    *to_root = p;
            return;
        }
        else if (cmp < 0)
        {
            ppar = p;
            p = p->left;
        }
        else
        {
            ppar = p;
            p = p->right;
        }
    }

    if ((new_node = (countnode*)malloc(sizeof(countnode))) == NULL)
    {
        fprintf(stderr,">E malloc failed in add_one()\n");
        exit(1);
    }

    if ((new_node->fmt = (char*)malloc(strlen(fmt)+1)) == NULL)
    {
        fprintf(stderr,">E malloc failed in add_one()\n");
        exit(1);
    }

    new_node->count = 1;
    strcpy(new_node->fmt,fmt);
    for (i = 0; i < nval; ++i)
	new_node->total[i] = val[i];

    new_node->left = new_node->right = NULL;

    if (cmp == 0)
    {
        *to_root = new_node;
	new_node->parent = NULL;
    }
    else if (cmp < 0)
    {
        ppar->left = new_node;
	new_node->parent = ppar;
    }
    else
    {
        ppar->right = new_node;
	new_node->parent = ppar;
    }

    splay(new_node);
    *to_root = new_node;
}

/****************************************************************************/

static int
scanline(char *s, char *f, number *val, int *valtype, char *outf)
/* Perform sscanf-like scan of line.
   The whole format must match.  outf is set to be an output format
   with unassigned values replaced by '*' and %s replaced by what
   it matches.  Assigned values except %s are put into val[] with
   their types in valtype[].  The number of them is returned.  If
   there is no match, -1 is returned.  See the comments at the start
   of the program for more information.
*/
{
	int n;  		 /* Number of values assigned */
	int fracdigits,digit;
	boolean doass,neednonsp;
	integer ival;
	double dval,digval;
	char ends;
	
	n = 0;

	while (*f != '\0')
	{
	    if (*f == '%')
	    {
		++f;
		if (*f == '*')
		{
		    doass = FALSE;
		    ++f;
		}
		else
		    doass = TRUE;

		if (*f == '%')
		{
		    if (!doass)
		    {
			fprintf(stderr,"Bad format item %%*\n");
			exit(1);
		    }
		    if (*s++ != '%') return -1;
		    ++f;
		    *outf++ = '%';
		    *outf++ = '%';
		}
	        else if (*f == 'c')
                {   
		    if (*s == ' ' || *s == '\t' || *s == '\n') return -1;
                    if (doass) *outf++ = *s;
		    else       *outf++ = '*';
                    ++f;
                    ++s;
                }
		else if (*f == 's')
		{
		    ends = *(f+1);
		    if (ends == ' ')
		    {
			while (*s == ' ' || *s == '\t')
			{
			    if (doass) *outf++ = *s;
                            ++s;
                        }
		    }
		    while (*s != '\n' && *s != ends)
		    {
			if (doass) *outf++ = *s;
			++s;
                    }
		    if (!doass) *outf++ = '*';
		    ++f;
		}
		else if (*f == 'd')
		{
		    while (*s == ' ' || *s == '\t') ++s;
		    if (!isdigit(*s)) return -1;
		    ival = 0;
		    while (isdigit(*s))
		    {
			digit =  *s++ - '0';
			if (ival > (maxint-digit)/10)
			{
			    fprintf(stderr,">E input integer too large\n");
			    exit(1);
			}
			ival = ival*10 + digit;
		    }
		    if (doass)
		    {
			valtype[n] = D;
			val[n++].d = ival;
			*outf++ = '%';
			*outf++ = 'd';
		    }
		    else
			*outf++ = '*';
		    ++f;
		}
		else if (*f == 'f')
		{
		    while (*s == ' ' || *s == '\t') ++s;
		    if (!isdigit(*s) && *s != '.') return -1;
		    dval = 0.0;
		    while (isdigit(*s)) dval = dval*10.0 + (*s++ - '0');
		    if (*s == '.')
		    {
			digval = 1.0;
			++s;
			while (isdigit(*s))
			{
			    digval /= 10.0;
			    dval += (*s++ - '0') * digval;
			}
		    }
		    if (doass)
                    {
			valtype[n] = F;
                        val[n++].f = dval;
                        *outf++ = '%';
                        *outf++ = 'f';
                    }
                    else
                        *outf++ = '*';
		    ++f;
		}
		else if (*f == ' ')
		{
		    while (*s == ' ' || *s == '\t') ++s;
		    *outf++ = ' ';
		    ++f;
		}
		else
		{
		    fprintf(stderr,"Bad format item %%%c\n",*f);
		    exit(1);
		}
	    }
	    else
	    {
		if (*s != *f) return -1;
		*outf++ = *f;
		++s;
		++f;
	    }
	}

	if (*s != '\0') return -1;

	*outf = '\0';
	return n;
}

/****************************************************************************/

static char*
makeformat(char *s)
/* Dynamically allocate a copy of s with the following changes:
   1.  If the last characters are \ and n, replace them by \n.
   2.  If the last character is not then \n, append \n.
*/
{
	char *news;
	size_t len;

	len = strlen(s);
	if ((news = (char*)malloc(len+2)) == NULL)
	{
	    fprintf(stderr,">E malloc failed\n");
	    exit(1);
	}

	strcpy(news,s);

	if (len >= 2 && news[len-1] == 'n' && news[len-2] == '\\')
	{
	    news[len-2] = '\n';
	    news[len-1] = '\0';
	}
	else if (len == 0 || news[len-1] != '\n')
	{
	    news[len] = '\n';
	    news[len+1] = '\0';
	}

	return news;
}

/****************************************************************************/

find_maxint()
{
/* Put the maximum possible integer value into maxint. */
	integer x;

	x = 1;
	while (x > 0) x <<= 1;
	x -= 1;

	if (x <= 0)
	{
	    fprintf(stderr,">E find_maxint() failed\n");
	    exit(1);
	}

	maxint = x;
}

/****************************************************************************/

main(int argc, char *argv[])
{
	int i,j,nvals,argnum;
	number val[MAXVALUES];
	int valtype[MAXVALUES];
	char line[MAXLINELEN+2];
	char outf[MAXLINELEN+MAXVALUES+6];
	long unmatched;
	int matched,finalmatched,errorlines;
	char *filename;
	FILE *infile;
	int numformats,firstarg;
	boolean havefinal,replaced;

	find_maxint();

	numformats = NUMFORMATS;
	firstarg = 1;

	if (numformats > MAXFORMATS)
	{
	    fprintf(stderr,">E increase MAXFORMATS\n");
	    exit(1);
	}

	replaced = FALSE;
	for (; firstarg < argc; ++firstarg)
	{
	    if (argv[firstarg][0] == '-' && argv[firstarg][1] != '\0')
	    {
		if (replaced)
		{
		    if (numformats == MAXFORMATS)
                    {
                        fprintf(stderr,">E increase MAXFORMATS\n");
                        exit(1);
                    }
		}
	        else
		    numformats = 0;
		format[numformats].flags = 0;
		format[numformats].fmt = makeformat(argv[firstarg]+1);
		++numformats;
		replaced = TRUE;
	    }
	    else if (argv[firstarg][0] == '+')
	    {
		if (numformats == MAXFORMATS)
        	{   
                    fprintf(stderr,">E increase MAXFORMATS\n");
                    exit(1);
                }
		format[numformats].flags = 0;
		format[numformats].fmt = makeformat(argv[firstarg]+1);
		++numformats;
	    }
	    else
		break;
	}

	havefinal = FALSE;
	for (i = 0; i < numformats; ++i)
	{
	    count_root[i] = NULL;
	    matching_lines[i] = 0;
	    if (HAS(i,FINAL)) havefinal = TRUE;
	}

	for (argnum = firstarg;
             argnum < (argc == firstarg ? argc+1 : argc); ++argnum)
	{
	    matched = finalmatched = errorlines = 0;
	    if (argnum >= argc || strcmp(argv[argnum],"-") == 0)
	    {
		infile = stdin;
		filename = "stdin";
	    }
	    else
	    {
		filename = argv[argnum];
		if ((infile = fopen(filename,"r")) == NULL)
	        {
		    fprintf(stderr,">E Can't open %s\n",filename);
		    exit(1);
		}
	    }

	    while (fgets(line,MAXLINELEN,infile) != NULL)
	    {
		line[MAXLINELEN] = '\n';
		line[MAXLINELEN+1] = '\0';
  	        if (line[0] == '\n') continue;
    
	        for (i = 0; i < numformats; ++i)
	        {
		    nvals = scanline(line,format[i].fmt,val,valtype,outf);
		    if (nvals >= 0)
		    {
			++matched;
			if (HAS(i,FINAL)) ++finalmatched;
			if (HAS(i,ERROR)) ++errorlines;
			++matching_lines[i];
		        add_one(&count_root[i],outf,nvals,val,valtype,i);
		        if (!HAS(i,CONTINUE)) break;
		    }
	        }
        
	        if (i == numformats) ++unmatched;
	    }
	    if (errorlines != 0)
	        printf("WARNING: Error lines in file %s\n",filename);
	    else if (matched == 0)
	        printf("WARNING: No matching lines in file %s\n",filename);
	    else if (finalmatched == 0 && havefinal)
	        printf("WARNING: No final lines in file %s\n",filename);
	    if (infile != stdin) fclose(infile);
	}

	for (i = 0; i < numformats; ++i)
	{
	    if (HAS(i,COUNT))
	    {
		if (matching_lines[i] > 0)
		    printf("%5ld lines matched ",matching_lines[i]);
		print_common(count_root[i]);
	    }
	    else
	        print_counts(count_root[i]);
	}

	if (unmatched > 0)
	    printf("%5ld non-empty lines not matched\n",unmatched);
	if (argc > firstarg)
	    printf("%5d files read altogether\n",argc-firstarg);
	if (errorlines > 0) 
	    printf("%5d errors found\n",errorlines);
}
